import googlemaps

from . import Geocode, Geotype, Location, Point, Viewport


class Client(googlemaps.Client):
    """Extend the Google Maps Client to format the response in a more usable format, which only
    includes data we care about, namely the viewport and the location type.
    """
    def geocode(self, search, country=None):
        """Force the search query to use the ISO country code as a component filter, to ensure the
        results are biased towards the country we know we are looking in.

        Returns a Geocode instance.
        """
        try:
            if country:
                response = super(Client, self).geocode(search, components={'country': country})
            else:
                response = super(Client, self).geocode(search)
        except googlemaps.exceptions.TransportError:
            # This exception is known to be raised when the connection is severed, which will
            # happen if the publication takes too long and the job gets killed.
            # When that happens, we just return None and move on, silencing the error.
            return None

        if not response:
            return None

        result = response[0]
        geotype = Geotype[result['geometry']['location_type']]
        southwest = Point(
            latitude=result['geometry']['viewport']['southwest']['lat'],
            longitude=result['geometry']['viewport']['southwest']['lng']
        )
        northeast = Point(
            latitude=result['geometry']['viewport']['northeast']['lat'],
            longitude=result['geometry']['viewport']['northeast']['lng']
        )
        viewport = Viewport(southwest=southwest, northeast=northeast)

        return Geocode(location=result.get('formatted_address', 'N/A'),
                       location_type=geotype,
                       viewport=viewport)

    def reverse_geocode(self, latlon):
        """Reverse-geocode the given latlon pair.

        The pair is provided as a (lat, lon) tuple.

        Returns a Geocode instance.
        Raise ValueError on failure.
        """
        try:
            response = super(Client, self).reverse_geocode(latlon)
        except googlemaps.exceptions.TransportError:
            # This exception is known to be raised when the connection is severed, which will
            # happen if the publication takes too long and the job gets killed.
            # When that happens, we just return None and move on, silencing the error.
            return None

        if not response:
            return None

        try:
            area = self._lookup('administrative_area_level_2', response)['long_name']
            country = self._lookup('country', response)['short_name']
            postal_code = self._lookup('postal_code', response)['long_name']
            postal_town = self._lookup('postal_town', response)['long_name']
        except ValueError as e:
            raise ValueError(f'Failed to resolve {latlon} - "{e}"')

        return Location(city=postal_town,
                        country=country,
                        formatted_address=response[0]['formatted_address'],
                        place_id=response[0]['place_id'],
                        postcode=postal_code,
                        region=area)

    def _lookup(self, field, response):
        """Loop through all addresses in the response to find the field we are after.

        When all fails, raise a ValueError.
        """
        for result in response:
            try:
                return [
                    address_component for address_component in result['address_components']
                    if field in address_component['types']
                ][0]
            except IndexError:
                continue

        raise ValueError(field)



"""Common RBX exceptions."""


class FatalException(Exception):
    """Raise this exception for non-transient errors.

    These exceptions can be given an extra details object, which may be used by the caller to log
    this as extra information.
    """
    def __init__(self, message='Something went wrong!', details=None):
        super().__init__(message)
        self.details = details

    def to_dict(self):
        rv = {
            'message': str(self)
        }

        if self.details:
            rv['details'] = self.details

        return rv


class Invalid(FatalException):
    """Generic invalid exception."""


class InvalidRequest(Invalid):
    """An Invalid exception with an HTTP status code and a URL."""
    def __init__(self, message, status_code=400, url=None, **kwargs):
        super().__init__(message, **kwargs)
        self.status_code = status_code
        self.url = url

    def to_dict(self):
        rv = super().to_dict()
        if self.url:
            rv['url'] = self.url

        return rv


class ClientException(InvalidRequest):
    """Raised from within the rbx.clients package."""


class InvalidVersion(FatalException):
    """Generic invalid version exception."""


class TransientException(Exception):
    """Raise this exception when we know the cause is transient (e.g.: connection errors,
    consistency error, ...).
    """



import click

from .maps import Client
from .packer import LocationPacker


@click.command(help='Fetch the Geocode for the given location.')
@click.argument('location')
@click.option('-c', '--country', help='Optional country code.')
@click.option('--key', help='google Cloud API key.', required=True)
def geocode(location, country=None, key=None):
    client = Client(key=key)
    geocode = client.geocode(location, country)

    if geocode is None:
        click.echo("Maps can't find '{}'".format(location))

    else:
        click.echo(click.style(geocode.location, fg='green'))
        click.echo('{}: {} ({})'.format(
            click.style('Precision', bold=True),
            geocode.location_type.name,
            geocode.location_type.value
        ))
        click.echo('{}: {}, {}'.format(
            click.style(' > Southwest', bold=True),
            geocode.viewport.southwest[0],
            geocode.viewport.southwest[1]
        ))
        click.echo('{}: {}, {}'.format(
            click.style(' > Northeast', bold=True),
            geocode.viewport.northeast[0],
            geocode.viewport.northeast[1]
        ))


@click.command(help='Fetch the Location the given lat/lon.')
@click.argument('latlon')
@click.option('--key', help='google Cloud API key.', required=True)
def reverse_geocode(latlon, key=None):
    client = Client(key=key)
    location = client.reverse_geocode(latlon)

    if location is None:
        click.echo("Maps can't find '{}'".format(latlon))

    else:
        click.echo(click.style(location.formatted_address, fg='green'))
        click.echo('{}: {}'.format(click.style('Country', bold=True), location.country))
        click.echo('{}: {}'.format(click.style('Region', bold=True), location.region))
        click.echo('{}: {}'.format(click.style('City', bold=True), location.city))
        click.echo('{}: {}'.format(click.style('Postcode', bold=True), location.postcode))
        click.echo('{}: {}'.format(click.style('Place ID', bold=True), location.place_id))


@click.command(help='Unpack the GeoIP databases from the DB_PATH folder.')
@click.argument('db_path')
@click.option('-d', '--data-dir', default='data',
              help='Destination folder for the extracted data (defaults to "./data").')
@click.option('-v', '--version', default='latest',
              help='GeoIP databases version (defaults to "latest").')
def unpack(db_path, data_dir, version):
    packer = LocationPacker(db_path=db_path, data_dir=data_dir, version=version)
    click.echo(f'Unpacking GeoIP2 databases to "{packer.data_dir}".')

    try:
        packer()
    except SystemExit:
        click.echo(click.style('Cannot find the database files.', fg='red'))
        raise SystemExit(1)
    else:
        click.echo('\U0001F44D Done.')
        click.echo(
            f'Now run "aws s3 sync {packer.data_dir} s3://<bucket>/rig/geo/{version}"'
            ' to upload to S3.')



import logging

from . import Client, HttpAuth

logger = logging.getLogger('rbx.clients.panels')


class PanelsClient(Client):
    """API Client for the Panels Service."""
    def __init__(self, endpoint, token):
        super().__init__()
        self.ENDPOINT = endpoint.rstrip('/')
        self.token = token

    @property
    def auth(self):
        """The Panels Service uses Digest Authentication."""
        return HttpAuth(self.token, key='X-RBX-TOKEN')

    def search(self, **parameters):
        """Perform a Panels search."""
        return self._post('/search', data=parameters)

    def set_weather_watch(self, caller, country, owner):
        """Set a watch on the specified country and owner, using the caller ID."""
        return self.request(
            method='get',
            path='/forecast/watch',
            data={
                'caller': caller,
                'country': country,
                'owner': owner
            }
        )

    def set_weather_unwatch(self, caller, country, owner):
        """Remove a watch on the specified country and owner, using the caller ID."""
        return self.request(
            method='get',
            path='/forecast/unwatch',
            data={
                'caller': caller,
                'country': country,
                'owner': owner
            }
        )



"""Common utility methods."""
import re
from urllib.parse import urljoin


def smart_url(url, force_https=False, prefix=None):
    """Return the appropriate URL.

    The returned URL will be:
     - an empty string, when the provided url is None;
     - the provided url, unchanged, when it includes the protocol ('http', 'https' or '//');
     - otherwise, the provided url, optionally prefixed with the given prefix.
     - finally, if force_https is set, we ensure the URL is return with the HTTPS protocol.
    """
    if not url:
        return ''

    if not force_https and (url.startswith('http') or url.startswith('//')):
        return url

    if prefix:
        url = urljoin(prefix, url)

    if force_https:
        match = re.match('http:|https:', url)
        if match:
            _, _, url = url.partition(match.group(0))

        url = urljoin('https://', url)

    return url



import logging

from . import Client, HttpAuth

logger = logging.getLogger('rbx.clients.reporting')


class ReportingClient(Client):
    """API Client for the Reporting Service."""
    def __init__(self, endpoint, token):
        super().__init__()
        self.ENDPOINT = endpoint.rstrip('/')
        self.token = token

    @property
    def auth(self):
        """The Reporting Service uses Digest Authentication."""
        return HttpAuth(self.token, key='X-RBX-TOKEN')

    def create_schedule(self, body, campaign, cron, format, name, recipients, reports):
        """Create a scheduled job.

        Return the job_id of the job just created.
        """
        response = self._post('/schedule', data={
            'body': body,
            'campaign': campaign,
            'cron': cron,
            'format': format,
            'name': name,
            'recipients': recipients,
            'reports': reports,
        })
        return response.get('job_id')

    def delete_schedule(self, job_id):
        """Create the scheduled job."""
        response = self._post('/schedule/delete', data={'job_id': job_id})
        return response.get('job_id')

    def pause_schedule(self, job_id):
        """Pause the scheduled job."""
        response = self._post('/schedule/pause', data={'job_id': job_id})
        return response.get('job_id')

    def resume_schedule(self, job_id):
        """Resume the scheduled job."""
        response = self._post('/schedule/resume', data={'job_id': job_id})
        return response.get('job_id')

    def update_schedule(self, body, campaign, cron, format, job_id, name, recipients, reports):
        """Update the scheduled job."""
        response = self._post('/schedule', data={
            'body': body,
            'campaign': campaign,
            'cron': cron,
            'format': format,
            'job_id': job_id,
            'name': name,
            'recipients': recipients,
            'reports': reports,
        })
        return response.get('job_id')



# Copyright 2020 Rockabox Media Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import find_namespace_packages, setup

setup(
    name='rbx',
    version='1.5.0',
    license='Apache 2.0',
    description='Scoota Platform utilities',
    long_description='A collection of common tools for the Scoota Platform services.',
    url='http://scoota.com/',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet',
    ],
    author='The Scoota Engineering Team',
    author_email='engineering@scoota.com',
    python_requires='>=3.6',
    install_requires=[
        'arrow>=0.15',
        'Click<8',
        'colorama',
        'PyYAML>=5.1.2',
    ],
    extras_require={
        'geo': [
            'googlemaps==4.2.0',
        ],
        'storage': [
            'google-cloud-storage==1.23',
        ],
        'tests': [
            'Flask==1.1.1',
            'googlemaps==4.2.0',
            'google-cloud-storage==1.23',
        ],
    },
    entry_points={
        'console_scripts': [
            'geocode = rbx.geo.cli:geocode [geo]',
            'reverse_geocode = rbx.geo.cli:reverse_geocode [geo]',
            'unpack = rbx.geo.cli:unpack [geo]',
        ],
    },
    packages=find_namespace_packages(include=['rbx.*']),
    zip_safe=True
)



import logging

import arrow

from . import Client, ClientException, HttpAuth

logger = logging.getLogger('rbx.clients.adsquare')


class AdsquareClient(Client):
    """Client for adsquare API.

    The API documentation can be found on the Platform wiki:
    https://github.com/rockabox/rbx/wiki/adsquare

    """
    AUTH_PATH = 'auth/login'
    ENDPOINT = 'https://amp.adsquare.com/api/v1/'
    STANDARD_COMPANY_ID = '588b6428e4b0d7d5f6895bd1'
    TIME_FORMAT = 'YYYY-MM-DDTHH:mm:ss.SSS[Z]'

    def __init__(self, company_id, dsp_id):
        super().__init__()
        self.company_id = company_id
        self.dsp_id = dsp_id

    @property
    def auth(self):
        """adsquare uses Digest Authentication."""
        return HttpAuth(self.token, key='X-AUTH-TOKEN')

    def create_config(self, country_code, name, start_date, end_date):
        """Create a new observation configuration.

        The start and end date are expected to be in ISO 8601, though any format understood by
        Arrow will work.
        """
        path = f'measurement/configs/byCompany/{self.company_id}'
        response = self.request('post', path, data={
            'dspId': self.dsp_id,
            'countryCode': country_code,
            'name': name,
            'active': True,
            'observationMode': 'full',
            'radiusInMeter': 50,
            'validFrom': arrow.get(start_date).format(self.TIME_FORMAT),
            'validTo': arrow.get(end_date).format(self.TIME_FORMAT),
        })
        return response['id']

    def disable_webhook(self, config_id):
        """Disable the registered webhook for the given configuration ID."""
        self.register_webhook(config_id, disable=True)

    def list_segments(self):
        """List all segments available.

        According to the documentation, there is no pagination on this endpoint, so we expect the
        full list to be returned with a single call.

        See: https://docs.adsquare.com/#!/enrichmentMetaApi/getEnrichmentAudiences_1
        """
        path = f'enrichmentMeta/audiences/{self.dsp_id}'
        response = self.request('get', path, data={'namev2': 'true'})

        return [{
            'company_id': audience['companyId'],
            'country': audience['countryCode'],
            'cpm': audience['cpm'],
            'currency': audience['currency'],
            'id': audience['audienceId'],
            'name': audience['name'],
            'owner': 'Standard' if audience['companyId'] == self.STANDARD_COMPANY_ID else 'Custom',
            'type': audience['type'],
        } for audience in response['audiences']]

    def pause_config(self, config_id):
        """Pause the given configuration."""
        self.update_config(config_id=config_id, active=False)

    def register_webhook(self, config_id, disable=False, feedback_url=None):
        """Register a Feedback URL webhook with adsquare for the given configuration ID.

        The same method can be used to re-enable an existing/disabled webhook.
        """
        data = {
            'enabled': not disable,
            'method': 'post',
            'contentType': 'csv',
        }
        if feedback_url:
            data['feedbackURL'] = feedback_url

        if not disable and not feedback_url:
            raise ClientException('Cannot register a webhook without a URL')

        return self.request('post', f'measurement/configs/{config_id}/feedback', data=data)

    def resume_config(self, config_id):
        """Resume the given configuration."""
        self.update_config(config_id=config_id, active=True)

    def update_config(self, config_id, active=None, start_date=None, end_date=None):
        """Update the given configuration ID.

        Only active and start/end dates can be changed.
        """
        data = {}

        if active is not None:
            data['active'] = active

        if start_date:
            data['validFrom'] = arrow.get(start_date).format(self.TIME_FORMAT)

        if end_date:
            data['validTo'] = arrow.get(end_date).format(self.TIME_FORMAT)

        return self.request('post', f'measurement/configs/{config_id}', data=data)

    def upload_config(self, config_id, points):
        """Upload a list of site locations to the given configuration.

        The points is expected to be a list of (lat, long) tuples as floating values.
        """
        data = ['id,lon,lat']
        for i, point in enumerate(points):
            data.append(f'{i},{point[1]},{point[0]}')

        # Make sure the payload has a trailing newline
        data.append('')

        self.request('put', f'measurement/configs/{config_id}/dataset/poi', data='\n'.join(data))



import logging
from urllib.parse import quote, parse_qs, urlparse

import arrow

from ..utils import smart_url
from . import Client, HttpAuth

logger = logging.getLogger('rbx.clients.broadsign')


class BroadsignClient(Client):
    """Client for Broadsign Reach APIs.

    The API documentation can be found on Google Drive:
    https://drive.google.com/file/d/16qS-xkOl5fcJaB26bf7STKn4qxkaZ6Mx/view

    """
    AUTH_PATH = '/auth/oauth2/token/'
    ENDPOINT = 'https://servssp.broadsign.com/'
    IF_MODIFIED_SINCE_FORMAT = 'ddd, DD MMM YYYY HH:mm:ss [GMT]'
    TOKEN = 'access_token'

    def __init__(self, dsp_id, dsp_name):
        super().__init__()
        self.csrftoken = None
        self.dsp_id = dsp_id
        self.dsp_name = dsp_name

    @property
    def auth(self):
        """Broadsign uses Digest Authentication."""
        return HttpAuth(f'Bearer {self.token}', key='Authorization')

    def get_message(self, response):
        return response.text

    def create_advertiser(self, name):
        """Create an Advertiser object."""
        response = self.request('post', '/api/advertisers/', data={
            'name': name,
            'dsp': {
                'id': self.dsp_id,
                'name': self.dsp_name
            },
        })
        return {
            'id': response['id'],
            'name': response['name'],
        }

    def create_creative(self, advertiser, creative_id, category, creative_type, creative_url, name,
                        publishers):
        """Create a Creative object.

        Note that the 'original_url' requires the protocol, or the API request will fail.
        """
        response = self.request('post', '/api/entity_creatives/', data={
            'advertiser': advertiser,
            'external_id': str(creative_id),
            'iab_categories': [category],
            'name': name,
            'original_url': smart_url(creative_url, force_https=True),
            'publishers': publishers,
            'type': 'VideoCreative' if creative_type == 'video' else 'ImageUrlCreative',
        })

        return {
            'id': int(response['external_id']),
            'name': response['name'],
            'status': response['status'],
            'type': response['type'],
        }

    def get_advertiser(self, name):
        """Retrieve an Advertiser object by name."""
        path = '/api/advertisers/?name={}'.format(quote(name))
        response = self.request('get', path)
        if response.get('count', 0) > 0:
            return {
                'id': response['results'][0]['id'],
                'name': response['results'][0]['name'],
            }

    def get_category(self, code):
        """Retrieve a Category object by code."""
        path = '/api/categories/?code={}'.format(code)
        response = self.request('get', path)
        if response.get('count', 0) > 0:
            return {
                'id': response['results'][0]['id'],
                'name': response['results'][0]['name'],
            }

    def get_creative(self, creative_id, as_entity=False):
        """Retrieve a Creative object by ID.

        Passing the as_entity flag as True will return the object as returned by the system.
        """
        path = '/api/entity_creatives/?external_id={}'.format(creative_id)
        response = self.request('get', path)
        if response.get('count', 0) > 0:
            entity = response['results'][0]
            if as_entity:
                return entity

            return {
                'id': int(entity['external_id']),
                'name': entity['name'],
                'status': entity['status'],
                'type': entity['type'],
            }

    def get_publishers(self, ad_type=None, page=None):
        """Retrieve a list of Publisher objects.

        The list may be further refined by ad_type, one of 'html', 'static', or 'video'.
        The API does not allow filtering by query string, so we have to resort to filtering the
        final results ourselves.

        Note that when a publisher has no ad types defined (i.e.: 'allowed_ad_types': []), we
        include the publisher. According to Broadsign, this happens when the details on the
        publisher haven't made it to the API yet, so we should add them.
        """
        if ad_type:
            assert ad_type in ('html', 'static', 'video'), f'Invalid Ad Type: "{ad_type}"'

        path = '/api/publishers/'
        params = {'page_size': 100}
        if page:
            params['page'] = page

        publishers = []
        response = self.request('get', path, data=params)
        if response.get('count', 0) > 0:
            publishers.extend(response['results'])

        if response.get('next'):
            try:
                page = int(parse_qs(urlparse(response.get('next')).query)['page'][0])
            except (KeyError, ValueError):
                pass
            else:
                publishers.extend(self.get_publishers(page=page))

        if ad_type:
            publishers[:] = [
                publisher for publisher in publishers
                if not publisher['allowed_ad_types'] or ad_type in [
                    allowed_ad_type['name'].lower()
                    for allowed_ad_type in publisher['allowed_ad_types']
                ]
            ]

        publishers.sort(key=lambda x: x['name'])

        return publishers

    def get_screens(self, last_modified=None, page=None):
        """Retrieve all screens from Broadsign.

        The list may be further refined using the last_modified date, which will cause the API to
        return only those screens that were modified after the specified date.
        """
        path = '/api/screens/'
        params = {'page_size': 100}
        if page:
            params['page'] = page

        headers = None

        if last_modified:
            utc = arrow.get(last_modified).to('utc')
            headers = {'If-Modified-Since': utc.format(self.IF_MODIFIED_SINCE_FORMAT)}

        screens = []
        response = self.request('get', path, data=params, headers=headers)
        if response.get('count', 0) > 0:
            screens.extend(response['results'])

        if response.get('next'):
            try:
                page = int(parse_qs(urlparse(response.get('next')).query)['page'][0])
            except (KeyError, ValueError):
                pass
            else:
                screens.extend(self.get_screens(last_modified=last_modified, page=page))

        return screens

    def update_creative(self, entity_id, payload):
        """Update a Creative object.

        Note that updating requires the entity_id, which is the ID of the Creative in the supplier
        system.
        """
        self.request('put', f'/api/entity_creatives/{entity_id}/', data=payload)



import logging
import logging.config
import os

import click


def init(format='[%(asctime)s] [%(levelname)s] %(message)s'):
    LOGGING = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': format
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'verbose'
            },
        },
        'loggers': {
            '': {
                'level': os.getenv('LOG_LEVEL', 'ERROR'),
                'handlers': ['console'],
            },
            'rbx': {
                'level': os.getenv('LOG_LEVEL', 'ERROR'),
                'handlers': ['console'],
                'propagate': False,
            },
        },
    }

    logging.config.dictConfig(LOGGING)
    logging.getLogger('urllib3').propagate = False
    logging.getLogger('google').propagate = False


class Logger:
    """Implements the default logging methods."""
    def __init__(self, format='[%(asctime)s] [%(levelname)s] %(message)s', logger=None):
        if not logger:
            logger = logging

        self.format = format
        self.logger = logger
        self.setup_logging()

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(self.__class__.__name__)

    def __getattr__(self, name):
        """Delegate the basic logging methods to the logger."""
        if name in ('debug', 'info', 'warning', 'error', 'critical', 'exception'):
            return getattr(self.logger, name)

        raise AttributeError(f'{self.__class__.__name__!r} object has no attribute {name!r}')

    def setup_logging(self):
        init(format=self.format)

    def done(self):
        pass

    def log(self, message):
        self.logger.debug(message)

    def progress(self):
        pass


class ConsoleLogger(Logger):
    """Logs to the console with colours."""

    def info(self, message, *args, **kwargs):
        self.logger.info(click.style(str(message), bold=True), *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.logger.warning(click.style(str(message), fg='yellow'), *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.logger.error(click.style(str(message), fg='bright_red'), *args, **kwargs)

    def exception(self, message, *args, **kwargs):
        self.logger.exception(click.style(str(message), fg='bright_red'), *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        self.logger.critical(
            click.style(str(message), fg='bright_white', bg='red'), *args, **kwargs)

    def log(self, message):
        """Print to screen without a new line."""
        click.echo(message, nl=False)

    def progress(self):
        """Print a '.' to screens without a new line."""
        click.echo('.', nl=False)

    def done(self):
        """Print 'done' followed by a new line."""
        click.secho(' done', fg='green')
